#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Raptor dependency processor
#

import re
import os
import xml.sax
import xml.dom.minidom as dom

from Blocks.Packaging.ComponentBuilder import ComponentBuilder
from Blocks.Packaging.PackageWriter import Dependency
from Blocks.Packaging.Rules import Rules
from Blocks.Packaging.Logging import Logging

class DotDeeDependencyProcessor(object):
    '''
    Get dependency info from ".d" dependency files found under the build
    directory. Each .d file is read for a list of files used during the build.
    The owner package of each of these is added as a dependency to the component
    source package.

    When calling dotDeeDependencyProcessor(component, buildData, globalFileMap)
    the relative directories under which dotDee files can be found must be
    listed in buildData.dependencyData["DotDeeDependencyProcessor"]["ddirs"] and
    buildData.dependencyData["DotDeeDependencyProcessor"]["buildDir"] overrides
    the default path to the build directory (epoc32/build).

    If additional dependency rules are used, the processor needs to replicate
    ComponentBuilder name resolution, which requires the location to the same
    directives as used by ComponentBuilder.
    '''
    # reduce log spam by not duplicating messages
    notInFileMap = []

    stats = {}
    stats["totalFiles"] = 0
    stats["notFoundSources"] = 0
    stats["notFoundAdditionals"] = 0
    stats["uniqueSourceDeps"] = 0
    stats["startedAdditionals"] = 0
    stats["badAdditionalPlatform"] = {}
    stats["matchedAdditionals"] = 0
    stats["badAdditionalPackageName"] = 0
    stats["additionals"] = 0
    stats["uniqueAdditionals"] = 0

    def __init__(self, config):
        '''
        Create a processor with default rules. These can be overridden using
        the I{config} parameter.

        @param config: None or a dictionary that may contain any of the
        following key-value pairs::
            "additionalDependencyRules" - parsed additional rules.
            "additionalDependencyRulesPath" - path to additional rules config.
            "directives" - parsed package directives.
            "directivesPath" - path to package directives config.
            "dependencyFilter" - Only include dependencies if the path starts
                                 with this string.
        Parsed configurations take precedence over paths if both are present.

        Both filepaths and configurations are supported to provide speed
        (less parsing) and convenience.
        '''
        self.logger = Logging.getLogger("pfw.processors.dotdee")
        self.deps = {}
        self.pathPrefix = ""
        if not isinstance(config, dict):
            config = {}

        self.additionalDependencyRules = config.get("additionalDependencyRules") or \
            AdditionalDependencyRules(config.get("additionalDependencyRulesPath")).getRules() or \
            AdditionalDependencyRules().getRules()

        self.directives = config.get("directives") or \
            dom.parse(config.get("directivesPath") or Rules.packageDirectives())

        self.setDependencyFilter(config.get("dependencyFilter", ""))

    def dReader(self, arg, dirPath, nameList):
        '''
        Read .d files in dirPath, adding paths to a list of dependencies with
        associated variantPlatform and variantType. These are obtained from the
        path. For example::
            full path == epoc32/release/armv5/urel/foo.dll
            variantPlatform == "armv5"
            variantType == "urel"
        Results in a dependecy on foo.dll associated with armv5-urel.

        @param arg: tuple(baseDir, epocRoot, sourceRoot)
        @param dirPath: Current dir being walk()'ed
        @param nameList: List of files and directories in dirPath

        arg is a tuple containing
            - baseDir is the top dir where walk() starts, needed to extrapolate
            the relative path of each .d, which in turn determines
            variantPlatform and variantType
            - epocRoot
            - sourceRoot
        '''
        baseDir, epocRoot, sourceRoot = arg
        epocRoot = os.path.normpath(epocRoot)
        sourceRoot = os.path.normpath(sourceRoot)
        dFiles = [n for n in nameList if n.lower().endswith(".d")]
        if not dFiles:
            return
        relativeDir = dirPath[len(baseDir):]
        parts = relativeDir.split(os.sep)
        for dFile in [os.path.join(dirPath, f) for f in dFiles if os.path.isfile(os.path.join(dirPath, f))]:
            deps = []
            fh = open(dFile)
            depString = ""
            for line in fh.readlines():
                line = line.strip()
                if line.endswith("\\"):
                    depString += line.rstrip("\\")
                else:
                    depString += line
                    for filePath in depString.split():
                        filePath = os.path.normpath(filePath)
                        if not filePath or filePath.endswith(":"):
                            continue
                        rel = os.path.relpath(filePath, epocRoot)
                        if not ".." in rel:
                            deps.append(rel)
                        else:
                            self.logger.debug("Ignoring '%s' - not in target root (%s)" % (filePath, epocRoot))
                    depString = ""

            fh.close()
            if deps:
                if len(parts) > 3:
                    self.addDepList(deps, variantPlatform=parts[2], variantType=parts[3])
                else:
                    self.addDepList(deps, "noarch")

    def addDepList(self, fileList, variantPlatform, variantType=None):
        '''
        Store dependencies in dictionary of dictionaries,
        self.deps[variantPlatform][variantType]

        @param fileList: List of paths relative to epocRoot. The component is
        dependent on the owners of these files.
        @type fileList: List(String)
        @param variantPlatform: The variant platform (e.g. armv5)
        @type variantPlatform: String
        @param variantType: e.g. udeb or urel
        @type variantType: String
        '''
        if variantPlatform != "noarch" and variantType == None:
            raise ValueError("Must have variantType if variantPlatform is %s" % variantPlatform)
        if variantPlatform == "noarch":
            variantType = "default"
        if variantPlatform not in self.deps:
            self.deps[variantPlatform] = {}
        if variantType not in self.deps[variantPlatform]:
            self.deps[variantPlatform][variantType] = []
        self.deps[variantPlatform][variantType].extend([f for f in fileList if f not in self.deps[variantPlatform][variantType]])

    def getDepList(self, variantSpec=None, filter=""):
        '''
        Get all dependencies by default; variantSpec is a list of tuples
        (variantPlatform, variantType) that can be used to get only those
        dependencies that are associated with them. An additional filter can
        be used to get only dependencies matching the filter path prefix.

        @param variantSpec: Variant platform and type.
        @type variantSpec: Tuple(String,String)
        @param filter: A string that must a path must match to be considered a
        valid dependency.
        @type filter: String
        '''
        ret = []
        if variantSpec:
            for platform, vtype in variantSpec:
                if platform == "noarch":
                    vtype = "default"
                if platform in self.deps and vtype in self.deps[platform]:
                    ret.extend([d for d in self.deps[platform][vtype] if d not in ret and d.startswith(filter)])
        else:
            for type in self.deps.values():
                for deps in type.values():
                    for dep in deps:
                        if dep not in ret and dep.startswith(filter):
                            ret.append(dep)
        return ret

    def setDependencyFilter(self, pathPrefix):
        ''' Does the same job as filter in getDepList, but for __call__. '''
        self.pathPrefix = pathPrefix

    def __call__(self, component, buildData, globalFileMap):
        buildDir = os.path.join(buildData.dependencyData.get(self.__class__.__name__, {}).get("buildDir"))
        if not buildDir:
            assert os.path.isdir(buildData.getTargetRoot())
            buildDir = os.path.join(buildData.getTargetRoot(), "epoc32", "build")
            self.logger.info("Build directory not given, using %s" % buildDir)
        if not os.path.isdir(buildDir):
            self.logger.info("Build directory %s does not exist" % buildDir)
            return

        ddirs = buildData.dependencyData.get(self.__class__.__name__, {}).get("ddirs", [])
        if not ddirs:
            self.logger.debug("%s: no dotDee directories" % buildData.getComponentName())
            return

        for d in ddirs:
            ddir = os.path.join(buildDir, d)
            assert os.path.isdir(ddir), "DotDee directory %s does not exist" % ddir
            os.path.walk(ddir, self.dReader, (ddir, buildData.getTargetRoot(), buildData.getSourceRoot()))

        for variantPlatform, typeDict in self.deps.items():
            for variantType, depList in typeDict.items():
                if variantPlatform == "noarch":
                    variantType = None
                for dep in depList:
                    DotDeeDependencyProcessor.stats["totalFiles"] += 1
                    if dep.startswith(self.pathPrefix):
                        if component.getSourcePackage():
                            requiredPackageName = globalFileMap.getPackageName(dep)
                            if not requiredPackageName:
                                if dep not in DotDeeDependencyProcessor.notInFileMap:
                                    if dep.startswith("epoc32"):
                                        self.logger.info("Source package dependency not in filemap: %s"%dep)
                                    DotDeeDependencyProcessor.notInFileMap.append(dep)
                                    DotDeeDependencyProcessor.stats["notFoundSources"] += 1
                                continue
                            if component.getSourcePackage().addDependency(Dependency(requiredPackageName, type="strong")):
                                DotDeeDependencyProcessor.stats["uniqueSourceDeps"] += 1
                                self.logger.debug("%s now depends on %s"%(component.getSourcePackage().getName(), requiredPackageName))
                        if self.additionalDependencyRules:
                            self.doAdditionalDependencies(dep, variantPlatform, variantType, component, globalFileMap)

    def doAdditionalDependencies(self, requiredFile, variantPlatform, variantType, component, globalFileMap):
        DotDeeDependencyProcessor.stats["startedAdditionals"] += 1
        for pathRegex, matchVariantPlatform, packages in self.additionalDependencyRules:
            if matchVariantPlatform and matchVariantPlatform != variantPlatform:
                if variantPlatform not in DotDeeDependencyProcessor.stats["badAdditionalPlatform"]:
                    DotDeeDependencyProcessor.stats["badAdditionalPlatform"][variantPlatform] = 1
                else:
                    DotDeeDependencyProcessor.stats["badAdditionalPlatform"][variantPlatform] += 1
                continue
            match = pathRegex.match(requiredFile)
            if match:
                DotDeeDependencyProcessor.stats["matchedAdditionals"] += 1
                # If component requires a matching file, the member package
                # identified by id also requires listed files.
                for id, files in packages: # packages are rulesets enclosed in <package> tags in the config
                    pkgKey = id.replace("$(VARIANTPLATFORM)", variantPlatform)
                    # I{pkgKey} is the raw name of the package depending on the
                    # first matching i{file} tag.
                    # By the time dependencyProcessors are being run, the
                    # original compname.variantplatform name has been lost
                    # so we must resolve the actual package name from pkgKey.
                    pkgName = ComponentBuilder.resolvePackageName(pkgKey, component.getName(), self.directives)
                    if pkgName == None:
                        self.logger.warning("%s: Unable to resolve package name from %s", component.getName(), pkgKey)
                        DotDeeDependencyProcessor.stats["badAdditionalPackageName"] += 1
                        continue
                    for file in files:
                        if "$(VARIANTTYPE)" in file and variantType == None:
                            self.logger.warning("Required placeholder VARIANTTYPE is None. Path: %s Filetag: %s", requiredFile, file)
                            continue
                        file = file.replace("$(VARIANTPLATFORM)", variantPlatform).replace("$(VARIANTTYPE)", variantType)
                        for matchId in range(len(match.groups())):
                            token = "$" + str(matchId + 1)
                            file = file.replace(token, match.groups()[matchId])
                        if file in DotDeeDependencyProcessor.notInFileMap:
                            continue
                        fileOwner = globalFileMap.getPackageName(file)
                        if not fileOwner:
                            self.logger.debug("Additional file %s not in filemap"%file)
                            DotDeeDependencyProcessor.notInFileMap.append(file)
                            DotDeeDependencyProcessor.stats["notFoundAdditionals"] += 1
                            continue
                        DotDeeDependencyProcessor.stats["additionals"] += 1
                        if not component.getPackage(pkgName):
                            self.logger.warning("Additional dependency failed: component %s has no package %s"%(component.getName(), pkgName))
                            continue
                        if component.getPackage(pkgName).addDependency(Dependency(fileOwner)):
                            DotDeeDependencyProcessor.stats["uniqueAdditionals"] += 1
                            self.logger.debug("Additional: %s now depends on %s"%(pkgName, fileOwner))
                # A match -> stop processing
                return True


class AdditionalDependencyRules(object):
    '''
    Read rules from XML file.

    Usage::
        rules = AdditionalDependencyRules().getRules()

    To specify alternate location::
        rules = AdditionalDependencyRules("my/rules.xml").getRules()

    '''
    defaultRules = os.path.join(os.path.dirname(__file__), "additionalDependencyRules.xml")

    class ConfigReader(xml.sax.ContentHandler):
        '''
        Read the configuration and store the results in self.rules, which is a
        list of tuples::
            (
                pathRegex, # regex
                variantPlatform, # string
                packages # list of tuples
            )

        Each package in turn is a tuple::
            (
                id,  # string
                files # list of strings
            )

        The id may contain placeholders, currently supported:
            - $(VARIANTPLATFORM)
            - $(VARIANTTYPE)
            - $(n) # n is a positive integer

        Counter to the example in the architecture document there is no
        $(COMPONENT), it will be hard coded.
        '''
        def __init__(self):
            self.rules = []
            self.pathRegex = None
            self.variantPlatform = None
            self.packages = []
            self.id = ""
            self.files = []

        def characters(self, data):
            self.cdata.append(data)

        def startElement(self, tag, attributes):
            self.cdata = []

        def endElement(self, tag):
            self.cdString = "".join(self.cdata).strip()
            self.cdata = []
            if tag == "path":
                self.pathRegex = re.compile(self.cdString)
            elif tag == "variantPlatform":
                self.variantPlatform = self.cdString
            elif tag == "id":
                self.id = self.cdString
            elif tag == "file":
                self.files.append(self.cdString)
            elif tag == "package":
                self.packages.append((self.id, self.files))
                self.files = []
            elif tag == "dep":
                self.rules.append((self.pathRegex, self.variantPlatform, self.packages))
                self.packages = []

    def __init__(self, configuration=None):
        '''
        Read the rules.

        @param configuration: Filename or file-like object containing the XML rules
        @type configuration: Openable or readable
        '''
        if not configuration:
            configuration = AdditionalDependencyRules.defaultRules
        p = xml.sax.make_parser()
        i = AdditionalDependencyRules.ConfigReader()
        p.setContentHandler(i)
        if hasattr(configuration, "read"):
            f = configuration
        else:
            f = open(configuration)
        while True:
            data = f.read()
            if not data:
                break
            p.feed(data)
        p.close()
        self.rules = i.rules

    def getRules(self):
        return self.rules